@@ -3,16 +3,68 @@ and defines various simple operations on it. *)
 open Pdfutil
 open Pdfio
 
+
 (* Predicate on characters delimiting entities. *)
 let is_delimiter = function
   | '(' | ')' | '<' | '>' | '[' | ']' | '{' | '}' | '%' | '/' -> true
   | _ -> false
 
 (* Streams of binary data, byte-addressable, can either be in memory (Got) or
-still in an input channel (ToGet). *)
+still in an input channel (ToGet). It may have been decrypted or encrypted, but
+the actual calculations deferred: this is indicated by the crypt record
+element. *)
+
+type saved_encryption =
+  {from_get_encryption_values :
+     Pdfcryptprimitives.encryption * string * string * int32 * string *
+     string option * string option;
+   encrypt_metadata : bool;
+   perms : string}
+
+type deferred_encryption =
+  {crypt_type : Pdfcryptprimitives.encryption;
+   file_encryption_key : string option;
+   obj : int;
+   gen : int;
+   key : int array;
+   keylength : int;
+   r : int}
+
+let print_deferred_encryption e =
+  begin match e.crypt_type with
+  | Pdfcryptprimitives.ARC4 (a, b) -> Printf.printf "crypt_type = ARC4 (%i, %i)\n" a b
+  | Pdfcryptprimitives.AESV2 -> Printf.printf "crypt_type = AESV2\n"
+  | Pdfcryptprimitives.AESV3 x -> Printf.printf "crypt_type = AESV3 %b\n" x
+  end;
+  begin match e.file_encryption_key with
+  | None -> Printf.printf "file_encryption_key: None\n"
+  | Some _ -> Printf.printf "file_encryption_key: Some\n"
+  end;
+  Printf.printf "key = %s\n" (Pdfio.string_of_int_array e.key);
+  Printf.printf "obj = %i, gen = %i\n" e.obj e.gen;
+  Printf.printf "keylength = %i\n" e.keylength;
+  Printf.printf "r = %i\n" e.r
+
+type toget_crypt =
+  | NoChange
+  | ToDecrypt of deferred_encryption
+
+type toget =
+  {input : input;
+   position : int;
+   length : int;
+   crypt : toget_crypt}
+
+let toget ?(crypt = NoChange) input position length =
+  {input; position; length; crypt}
+
+let length_of_toget t = t.length
+let position_of_toget t = t.position
+let input_of_toget t = t.input
+
 type stream =
   | Got of bytes
-  | ToGet of input * int * int (*r input, position, length *)
+  | ToGet of toget
 
 (* Type for individual PDF objects. A Name includes the initial `/'. A
 Stream consists of a reference to a pair of the stream dictionary (another
@@ -35,6 +87,9 @@ type pdfobject =
   | Stream of (pdfobject * stream) ref
   | Indirect of int
 
+(* For debug. Filled in by Pdfwrite *)
+let string_of_pdf : (pdfobject -> string) ref = ref (function _ -> "") 
+
 (* An object is either lexed, or needs to be lexed from a position in the
 input.
   Parsed -
@@ -89,17 +144,6 @@ type pdfobjects =
    mutable pdfobjects : pdfobjmap;
    mutable object_stream_ids : (int, int) Hashtbl.t}
 
-type encryption = 
-  | ARC4 of int * int
-  | AESV2
-  | AESV3 of bool (* true = iso, false = old algorithm *)
-
-type saved_encryption =
-  {from_get_encryption_values :
-     encryption * string * string * int32 * string *
-     string option * string option;
-   encrypt_metadata : bool;
-   perms : string}
 
 (* PDF Document. The major and minor version numbers, the root object number,
 the list of objects and the trailer dictionary.
@@ -112,6 +156,7 @@ type t =
    mutable root : int;
    mutable objects : pdfobjects; 
    mutable trailerdict : pdfobject;
+   mutable was_linearized : bool;
    mutable saved_encryption : saved_encryption option}
 
 (* The null PDF document. *)
@@ -125,6 +170,7 @@ let empty () =
       pdfobjects = pdfobjmap_empty ();
       object_stream_ids = null_hash ()};
    trailerdict = Dictionary [];
+   was_linearized = false;
    saved_encryption = None}
 
 (* General exception for low-level errors. *)
@@ -144,26 +190,84 @@ let is_not_whitespace = function
   | '\000' | '\009' | '\010' | '\012' | ' ' | '\013' -> false
   | _ -> true
 
+let process_deferred_cryption toget_crypt data =
+  match toget_crypt with
+    NoChange -> data
+  | ToDecrypt saved ->
+      (*Printf.printf "Forcing...\n";
+      print_deferred_encryption saved;*)
+      Pdfcryptprimitives.decrypt_stream_data
+        saved.crypt_type
+        false
+        saved.file_encryption_key
+        saved.obj
+        saved.gen
+        saved.key
+        saved.keylength
+        saved.r
+        data
+
+let remove_string_compare (k' : string) l =
+  let rec remove_inner r (k' : string) = function
+    | [] -> r
+    | (k, _)::t when k = k' -> List.rev_append r t
+    | h::t -> remove_inner (h::r) k' t
+  in
+    remove_inner [] k' l
+
+(* Remove a dictionary entry. Also works for streams. *)
+let rec remove_dict_entry dict key =
+  match dict with
+  | Dictionary d -> Dictionary (remove_string_compare key d)
+  | Stream ({contents = (dict', stream)} as s) ->
+      s := (remove_dict_entry dict' key, stream);
+      Stream s
+  | _ -> raise (PDFError "remove_dict_entry: not a dictionary")
+
+(* Replace dict entry, raising Not_found if it's not there. Also works
+for streams. *)
+let rec replace_dict_entry dict key value =
+  match dict with
+  | Dictionary d -> Dictionary (replace key value d)
+  | Stream ({contents = (dict', stream)} as s) ->
+      s := (replace_dict_entry dict' key value, stream);
+      Stream s
+  | _ -> raise (PDFError "replace_dict_entry: not a dictionary.")
+
+(* Add a dict entry, replacing if there. Also works for streams. *)
+let rec add_dict_entry dict key value =
+  match dict with
+  | Dictionary d -> Dictionary (add key value d)
+  | Stream ({contents = (dict', stream)} as s) ->
+      s := (add_dict_entry dict' key value, stream);
+      Stream s
+  | _ -> raise (PDFError "add_dict_entry: not a dictionary.")
+
 (* Get a stream from disk if it hasn't already been got. *)
 let getstream = function
-  | Stream ({contents = (d, ToGet (i, o, l))} as stream) ->
+  | Stream ({contents = (d, ToGet {input = i; position = o; length = l; crypt})} as stream) ->
       if l = 0 then stream := (d, Got (mkbytes 0)) else
-        let s = mkbytes l in
-          begin try
-            i.seek_in o;
-            for c = 0 to l - 1 do
-              match i.input_byte () with
-              | b when b = Pdfio.no_more -> raise End_of_file
-              | b -> bset_unsafe s c b
-            done;
-            stream := (d, Got s)
-          with
-            End_of_file ->
-              raise (PDFError "Pdf.getstream: can't read stream.")
-          end
-  | Stream _ -> ()
+        begin try
+          let data =
+            process_deferred_cryption crypt (Pdfio.bytes_of_input i o l)
+          in
+            (* Correct the length *)
+            let d' =
+              match d with
+                Dictionary _ -> replace_dict_entry d "/Length" (Integer (bytes_size data))
+              | _ -> d (* May be null in Pdfwrite.WStream *)
+            in
+              stream := (d', Got data)
+        with
+          e ->
+            raise
+              (PDFError
+                ("Pdf.getstream: can't read stream" ^ Printexc.to_string e))
+        end
+  | Stream {contents = (_, Got _)} -> ()
   | _ -> raise (PDFError "Pdf.getstream: not a stream")
 
+
 let recurse_array (f : pdfobject -> pdfobject) elts =
   Array (map f elts)
 
@@ -724,41 +828,6 @@ let objects_referenced no_follow_entries no_follow_contains pdf pdfobject =
       no_follow_entries no_follow_contains pdf set 0 (Parsed pdfobject);
     refset_elts !set
 
-let remove_string_compare (k' : string) l =
-  let rec remove_inner r (k' : string) = function
-    | [] -> r
-    | (k, _)::t when k = k' -> List.rev_append r t
-    | h::t -> remove_inner (h::r) k' t
-  in
-    remove_inner [] k' l
-
-(* Remove a dictionary entry. Also works for streams. *)
-let rec remove_dict_entry dict key =
-  match dict with
-  | Dictionary d -> Dictionary (remove_string_compare key d)
-  | Stream ({contents = (dict', stream)} as s) ->
-      s := (remove_dict_entry dict' key, stream);
-      Stream s
-  | _ -> raise (PDFError "remove_dict_entry: not a dictionary")
-
-(* Replace dict entry, raising Not_found if it's not there. Also works
-for streams. *)
-let rec replace_dict_entry dict key value =
-  match dict with
-  | Dictionary d -> Dictionary (replace key value d)
-  | Stream ({contents = (dict', stream)} as s) ->
-      s := (replace_dict_entry dict' key value, stream);
-      Stream s
-  | _ -> raise (PDFError "replace_dict_entry: not a dictionary.")
-
-(* Add a dict entry, replacing if there. Also works for streams. *)
-let rec add_dict_entry dict key value =
-  match dict with
-  | Dictionary d -> Dictionary (add key value d)
-  | Stream ({contents = (dict', stream)} as s) ->
-      s := (add_dict_entry dict' key value, stream);
-      Stream s
-  | _ -> raise (PDFError "add_dict_entry: not a dictionary.")
 
 (* Find the contents of a stream as a bytes. *)
 let bigarray_of_stream s =
@@ -907,5 +976,7 @@ let deep_copy from =
       pdfobjects = deep_copy_pdfobjects from from.objects.pdfobjects;
       object_stream_ids = Hashtbl.copy from.objects.object_stream_ids};
    trailerdict = from.trailerdict;
+   was_linearized = from.was_linearized;
    saved_encryption = from.saved_encryption}
 
+
