@@ -4,24 +4,6 @@ open Pdfio
 
 let crypt_debug = ref false
 
-(* Given an object number, generation number, input key and key length in bits,
-apply Algorithm 3.1 from the PDF Reference manual to obtain the hash to be used
-by the encryption function. *)
-let find_hash crypt_type obj gen key keylength =
-  let from_obj =
-    [| i32toi (land32 obj 0x000000ffl);
-       i32toi (lsr32 (land32 obj 0x0000ff00l) 8);
-       i32toi (lsr32 (land32 obj 0x00ff0000l) 16) |]
-  in let from_gen =
-    [| i32toi (land32 gen 0x000000ffl);
-       i32toi (lsr32 (land32 gen 0x0000ff00l) 8) |]
-  in let extra =
-    if crypt_type = Pdf.AESV2 then [| 0x73; 0x41; 0x6C; 0x54 |] else [| |]
-  in
-    let digest_input = string_of_int_arrays [key; from_obj; from_gen; extra] in
-      int_array_of_string
-        (String.sub (Digest.string digest_input) 0 (min 16 (keylength / 8 + 5)))
-
 (* Find a key, given a password, O entry, P entry, id entry, and key length in
 bits. *)
 let paddings =
@@ -116,8 +98,12 @@ flprint "END_AUTHENTICATE_USER\n";*)
             (Pdfcryptprimitives.crypt key (bytes_of_int_array paddings))
 
 (* Decrypt a PDF file, given the user password. *)
+
+(* For debug. Filled in by Pdfwrite *)
 let string_of_pdf : (Pdf.pdfobject -> string) ref = ref (function _ -> "")
 
+
+
 let rec decrypt
   crypt_type pdf no_encrypt_metadata encrypt obj gen key keylength r
   file_encryption_key l
@@ -132,11 +118,11 @@ let rec decrypt
       | {contents = Pdf.ParsedAlreadyDecrypted _} -> Pdf.String s
       | _ (* Will always be Parsed for now...*) ->
         let f =
-          (if crypt_type = Pdf.AESV2 then
+          (if crypt_type = Pdfcryptprimitives.AESV2 then
             (if encrypt
                then Pdfcryptprimitives.aes_encrypt_data 4
                else Pdfcryptprimitives.aes_decrypt_data 4)
-           else if (match crypt_type with Pdf.AESV3 _ -> true | _ -> false) then
+           else if (match crypt_type with Pdfcryptprimitives.AESV3 _ -> true | _ -> false) then
             (if encrypt
                then Pdfcryptprimitives.aes_encrypt_data 8
                else Pdfcryptprimitives.aes_decrypt_data 8)
@@ -154,7 +140,7 @@ let rec decrypt
                   (string_of_bytes (f (int_array_of_string key) s_ints))
             else
               let hash =
-                find_hash crypt_type (i32ofi obj) (i32ofi gen) key keylength
+                Pdfcryptprimitives.find_hash crypt_type (i32ofi obj) (i32ofi gen) key keylength
               in
                 Pdf.String (string_of_bytes (f hash s_ints))
       end
@@ -186,56 +172,86 @@ let rec decrypt
       end
   | x -> x
 
+and is_identity no_encrypt_metadata pdf d =
+  let identity_crypt_filter_present =
+    match Pdf.lookup_direct pdf "/Filter" d with
+    | Some (Pdf.Name "/Crypt")
+    | Some (Pdf.Array (Pdf.Name "/Crypt"::_)) ->
+        begin match Pdf.lookup_direct pdf "/DecodeParms" d with
+        | Some (Pdf.Dictionary decodeparmsdict)
+        | Some (Pdf.Array (Pdf.Dictionary decodeparmsdict::_)) ->
+            begin match
+              Pdf.lookup_direct
+                pdf "/Name" (Pdf.Dictionary decodeparmsdict)
+            with
+            | Some (Pdf.Name "/Identity") | None -> true
+            | _ -> false
+            end
+        | _ -> true
+        end
+    | _ -> false
+  in
+    (no_encrypt_metadata &&
+       (match
+          Pdf.lookup_direct pdf "/Type" d with
+            Some (Pdf.Name "/Metadata") -> true
+          | _ -> false))
+    || identity_crypt_filter_present
+
 and decrypt_stream
   crypt_type pdf no_encrypt_metadata encrypt obj gen key keylength r
   file_encryption_key stream
 =
-  Pdf.getstream stream;
   begin match stream with
-  | Pdf.Stream {contents = (Pdf.Dictionary dict as d, Pdf.Got data)} ->
-      if
-        begin let identity_crypt_filter_present =
-          match Pdf.lookup_direct pdf "/Filter" d with
-          | Some (Pdf.Name "/Crypt")
-          | Some (Pdf.Array (Pdf.Name "/Crypt"::_)) ->
-              begin match Pdf.lookup_direct pdf "/DecodeParms" d with
-              | Some (Pdf.Dictionary decodeparmsdict)
-              | Some (Pdf.Array (Pdf.Dictionary decodeparmsdict::_)) ->
-                  begin match
-                    Pdf.lookup_direct
-                      pdf "/Name" (Pdf.Dictionary decodeparmsdict)
-                  with
-                  | Some (Pdf.Name "/Identity") | None -> true
-                  | _ -> false
-                  end
-              | _ -> true
-              end
-          | _ -> false
-        in
-          (no_encrypt_metadata &&
-             (match
-                Pdf.lookup_direct pdf "/Type" d with
-                  Some (Pdf.Name "/Metadata") -> true
-                | _ -> false))
-          || identity_crypt_filter_present
-        end
-      then
-        stream
-      else
+  | Pdf.Stream {contents = (Pdf.Dictionary dict as d, data)} ->
+      if is_identity no_encrypt_metadata pdf d then stream else
         let data' =
-          let f =
-            (if crypt_type = Pdf.AESV2 then
-               (if encrypt
-                  then Pdfcryptprimitives.aes_encrypt_data 4
-                  else Pdfcryptprimitives.aes_decrypt_data 4)
-             else if
-               (match crypt_type with Pdf.AESV3 _ -> true | _ -> false)
-             then
-               (if encrypt
-                  then Pdfcryptprimitives.aes_encrypt_data 8
-                  else Pdfcryptprimitives.aes_decrypt_data 8)
-             else
-               Pdfcryptprimitives.crypt)
+          let rec f unhashed_key key data =
+            let crypt = Pdf.ToDecrypt {Pdf.crypt_type; file_encryption_key; obj; gen; key = unhashed_key; keylength; r} in
+            match data with
+              Pdf.Got data ->
+                (*Printf.printf "decrypt_stream: Got, encrypt = %b\n" encrypt;*)
+                (if crypt_type = Pdfcryptprimitives.AESV2 then
+                   (if encrypt
+                      then
+                        Pdf.Got (Pdfcryptprimitives.aes_encrypt_data 4 key data)
+                      else
+                        Pdf.ToGet
+                          (Pdf.toget ~crypt (Pdfio.input_of_bytes data) 0 (bytes_size data)))
+                 else if
+                   (match crypt_type with Pdfcryptprimitives.AESV3 _ -> true | _ -> false)
+                 then
+                   (if encrypt
+                      then
+                        Pdf.Got (Pdfcryptprimitives.aes_encrypt_data 8 key data)
+                      else
+                        Pdf.ToGet
+                          (Pdf.toget ~crypt (Pdfio.input_of_bytes data) 0 (bytes_size data)))
+                 else
+                   if encrypt then
+                     Pdf.Got (Pdfcryptprimitives.crypt key data)
+                   else
+                     Pdf.ToGet
+                       (Pdf.toget ~crypt (Pdfio.input_of_bytes data) 0 (bytes_size data)))
+            | Pdf.ToGet toget ->
+                (*Printf.printf "decrypt_stream: ToGet, encrypt = %b\n" encrypt;*)
+                if encrypt then 
+                  (* If encrypting, call getstream, then go again *)
+                  begin
+                    Pdf.getstream stream;
+                    match stream with
+                      Pdf.Stream {contents = (_, data)} -> f unhashed_key key data
+                    | _ -> assert false
+                  end
+                else
+                  (* Otherwise, do the deferred decryption magic *)
+                  let crypt =
+                    Pdf.ToDecrypt
+                      {Pdf.crypt_type; file_encryption_key; obj; gen; key; keylength; r}
+                  in
+                    Pdf.ToGet
+                      (Pdf.toget ~crypt
+                        (Pdf.input_of_toget toget) (Pdf.position_of_toget toget) (Pdf.length_of_toget toget))
           in
             if r = 5 || r = 6 then
               let key =
@@ -243,12 +259,12 @@ and decrypt_stream
                   Some k -> k
                 | None -> raise (Pdf.PDFError "decrypt: no key C")
               in
-                f (int_array_of_string key) data
+                f (int_array_of_string key) (int_array_of_string key) data
             else
               let hash =
-                find_hash crypt_type (i32ofi obj) (i32ofi gen) key keylength
+                Pdfcryptprimitives.find_hash crypt_type (i32ofi obj) (i32ofi gen) key keylength
               in
-                f hash data
+                f key hash data
         in let dict' =
           Pdf.recurse_dict
             (decrypt crypt_type pdf no_encrypt_metadata encrypt
@@ -256,13 +272,12 @@ and decrypt_stream
             dict
         in
           let dict'' =
-            if bytes_size data <> bytes_size data' then
+            match data' with Pdf.Got data' ->
               Pdf.replace_dict_entry
                 dict' "/Length" (Pdf.Integer (bytes_size data'))
-            else
-              dict'
+            | _ -> dict'
           in
-            Pdf.Stream {contents = (dict'', Pdf.Got data')}
+            Pdf.Stream {contents = (dict'', data')}
   | _ -> assert false
   end
 
@@ -324,10 +339,10 @@ let get_encryption_values pdf =
         with
         | Some (Pdf.Name "/Standard"), Some (Pdf.Integer 1), _, Some (Pdf.Integer r)
         | Some (Pdf.Name "/Standard"), Some (Pdf.Integer 2), None, Some (Pdf.Integer r) ->
-            Some (Pdf.ARC4 (40, r))
+            Some (Pdfcryptprimitives.ARC4 (40, r))
         | Some (Pdf.Name "/Standard"), Some (Pdf.Integer 2), Some (Pdf.Integer n), _
             when n mod 8 = 0 && n >= 40 && n <= 128 ->
-              Some (Pdf.ARC4 (n, 3))
+              Some (Pdfcryptprimitives.ARC4 (n, 3))
         | Some (Pdf.Name "/Standard"), Some (Pdf.Integer (4 | 5)), length, Some (Pdf.Integer r) ->
             begin match Pdf.lookup_direct pdf "/CF" encryptdict with
             | Some cfdict ->
@@ -336,15 +351,15 @@ let get_encryption_values pdf =
                     begin match Pdf.lookup_direct pdf "/CFM" stdcfdict with
                     | Some (Pdf.Name "/V2") ->
                         begin match length with
-                        | Some (Pdf.Integer i) -> Some (Pdf.ARC4 (i, 4))
+                        | Some (Pdf.Integer i) -> Some (Pdfcryptprimitives.ARC4 (i, 4))
                         | _ ->
                             begin match Pdf.lookup_direct pdf "/Length" cfdict with
-                            | Some (Pdf.Integer i) -> Some (Pdf.ARC4 (i, 4))
+                            | Some (Pdf.Integer i) -> Some (Pdfcryptprimitives.ARC4 (i, 4))
                             | _ -> None
                             end
                         end
-                    | Some (Pdf.Name "/AESV2") -> Some Pdf.AESV2
-                    | Some (Pdf.Name "/AESV3") -> Some (Pdf.AESV3 (r = 6))
+                    | Some (Pdf.Name "/AESV2") -> Some Pdfcryptprimitives.AESV2
+                    | Some (Pdf.Name "/AESV3") -> Some (Pdfcryptprimitives.AESV3 (r = 6))
                     | _ -> None
                     end
                 | _ -> None
@@ -354,7 +369,7 @@ let get_encryption_values pdf =
         | _ -> None
       in
         let chop_string3248 n s =
-          let need = match crypt_type with Some (Pdf.AESV3 _) -> 48 | _ -> 32 in
+          let need = match crypt_type with Some (Pdfcryptprimitives.AESV3 _) -> 48 | _ -> 32 in
             if String.length s < need then
               raise (Pdf.PDFError (n ^ ": too small in get_encryption_values"))
             else
@@ -474,60 +489,52 @@ let mod3 b =
     !x mod 3
 
 let prs s =
-  String.iter (fun x -> Printf.printf "%02x" (int_of_char x)) (if String.length s > 16 then String.sub s 0 16 else s);
+  String.iter
+    (fun x -> Printf.printf "%02x" (int_of_char x))
+    (if String.length s > 16 then String.sub s 0 16 else s);
   flprint "\n"
 
-let shamix password udata i =
-  if !crypt_debug then
-    begin
-      flprint "Beginning of shamix\n";
-      flprint "Password is\n";
-      prs password;
-      flprint "udata is\n";
-      prs (match udata with None -> "" | Some x -> x);
-    end;
-  let k = ref (Pdfcryptprimitives.sha256 i)
-  and fin = ref false
-  and round = ref 0
-  and last_e = ref 0 in
-    (*flprint "initial hash is\n";
-    prs !k;*)
-    while not !fin do
-      round += 1;
-      (*Printf.printf "THIS IS ROUND %i\n" !round;*)
-      let k1 = password ^ !k ^ match udata with None -> "" | Some x -> x in
-        (*flprint "K1 is\n";
-        prs k1;*)
-        let k1_64 = String.concat "" (many k1 64) in
-          let e =
-            let key = int_array_of_string (String.sub !k 0 16) 
-            and firstblock = int_array_of_string (String.sub !k 16 16) in
-              let raw =
-                string_of_bytes (Pdfcryptprimitives.aes_encrypt_data ~firstblock:firstblock 4 key (Pdfio.bytes_of_string k1_64))
-              in
-                String.sub raw 16 (String.length raw - 32)
-          in
-            (*flprint "E is:\n";
-            prs e;*)
-            last_e := int_of_char e.[String.length e - 1];
-            k := (match mod3 e with
-                    0 -> (*flprint "using sha256\n";*) Pdfcryptprimitives.sha256
-                    | 1 -> (*flprint "using sha384\n";*) Pdfcryptprimitives.sha384
-                    | _ -> (*flprint "using sha512\n";*) Pdfcryptprimitives.sha512) (Pdfio.input_of_string e);
-            (*flprint "New k is:\n";
-            prs !k;
-      Printf.printf "!last_e = %i, !round - 32 = %i\n" !last_e (!round - 32);*)
-      fin := !round >= 64 && !last_e <= !round - 32
-    done;
-    let result =
-    String.sub !k 0 32
-    in
+let shamix_cache = Hashtbl.create 16
+
+let shamix password udata s =
+  try Hashtbl.find shamix_cache (password, udata, s) with
+    Not_found ->
+      let i = input_of_string s in
       if !crypt_debug then
         begin
-          flprint "RESULT:\n";
-          prs result
+          flprint "Beginning of shamix\n Password is\n"; prs password; flprint "udata is\n";
+          prs (match udata with None -> "" | Some x -> x);
         end;
-      result
+      let k = ref (Pdfcryptprimitives.sha256 i)
+      and fin = ref false
+      and round = ref 0
+      and last_e = ref 0 in
+        while not !fin do
+          round += 1;
+          let k1 = password ^ !k ^ match udata with None -> "" | Some x -> x in
+            let k1_64 = String.concat "" (many k1 64) in
+              let e =
+                let key = int_array_of_string (String.sub !k 0 16) 
+                and firstblock = int_array_of_string (String.sub !k 16 16) in
+                  let raw =
+                    string_of_bytes
+                      (Pdfcryptprimitives.aes_encrypt_data
+                         ~firstblock:firstblock 4 key (Pdfio.bytes_of_string k1_64))
+                  in
+                    String.sub raw 16 (String.length raw - 32)
+              in
+                last_e := int_of_char e.[String.length e - 1];
+                k := (match mod3 e with
+                          0 -> Pdfcryptprimitives.sha256
+                        | 1 -> Pdfcryptprimitives.sha384
+                        | _ -> Pdfcryptprimitives.sha512) (Pdfio.input_of_string e);
+          fin := !round >= 64 && !last_e <= !round - 32
+        done;
+        let result = String.sub !k 0 32 in
+          if !crypt_debug then
+            begin flprint "RESULT:\n"; prs result end;
+          Hashtbl.add shamix_cache (password, udata, s) result;
+          result
 
 (* Conversion via unicode and SASLprep required here for complicated passwords. *)
 let make_utf8 pw =
@@ -546,8 +553,9 @@ let file_encryption_key_aesv3 ?digest iso utf8pw o oe u =
       | None ->
           let i =
             int_array_of_string
-            ((if iso then shamix utf8pw (Some u) else Pdfcryptprimitives.sha256)
-                (Pdfio.input_of_string (String.concat "" [utf8pw; String.sub o 40 8; String.sub u 0 48])))
+            ((if iso then shamix utf8pw (Some u) else (fun x ->
+              Pdfcryptprimitives.sha256 (Pdfio.input_of_string x)))
+                (String.concat "" [utf8pw; String.sub o 40 8; String.sub u 0 48]))
           in
             if Array.length i <> 32 then Printf.printf "file_encryption_key_aesv3 made length %i\n" (Array.length i);
             i
@@ -558,7 +566,10 @@ let file_encryption_key_aesv3_user iso utf8pw u ue =
   if String.length u < 48 then raise (Pdf.PDFError "/U too short in file_encryption_key_aesv3_user") else
     Pdfcryptprimitives.aes_decrypt_data ~remove_padding:false
       8
-      (int_array_of_string ((if iso then shamix utf8pw None else Pdfcryptprimitives.sha256) (Pdfio.input_of_string (String.concat "" [utf8pw; String.sub u 40 8]))))
+      (int_array_of_string
+        ((if iso then shamix utf8pw None else (fun x ->
+          Pdfcryptprimitives.sha256 (Pdfio.input_of_string x)))
+         (String.concat "" [utf8pw; String.sub u 40 8])))
       (bytes_of_string (zero_iv ^ ue))
 
 (* Algorithm 3.12 - Authenticating the owner password. *)
@@ -566,15 +577,19 @@ let authenticate_owner_password_aesv3 iso utf8pw u o =
   if String.length o < 48 || String.length u < 48 then
     raise (Pdf.PDFError "/O too short in authenticate_owner_password")
   else
-      (if iso then shamix utf8pw (Some u) else Pdfcryptprimitives.sha256)
-      (Pdfio.input_of_string (String.concat "" [utf8pw; String.sub o 32 8; String.sub u 0 48]))
+      (if iso then shamix utf8pw (Some u) else (fun x ->
+        Pdfcryptprimitives.sha256 (Pdfio.input_of_string x)))
+      (String.concat "" [utf8pw; String.sub o 32 8; String.sub u 0 48])
     =
       String.sub o 0 32
 
 (* Algorithm 3.11 - Authenticating the user password. *)
 let authenticate_user_password_aesv3 iso utf8pw u =
   if String.length u < 48 then raise (Pdf.PDFError "/U too short in authenticate_owner_password") else
-    (if iso then shamix utf8pw None else Pdfcryptprimitives.sha256) (Pdfio.input_of_string (String.concat "" [utf8pw; String.sub u 32 8])) = String.sub u 0 32
+    (if iso then shamix utf8pw None else (fun x -> Pdfcryptprimitives.sha256
+    (Pdfio.input_of_string x)))
+    (String.concat "" [utf8pw; String.sub u 32 8])
+  = String.sub u 0 32
 
 (* Part of algorithm 3.2a - return p from perms so we can check they match *)
 let p_of_perms key perms =
@@ -597,8 +612,8 @@ let decrypt_pdf ?keyfromowner user_pw pdf =
      let crypt_type, u, o, p, id, ue, oe = get_encryption_values pdf in
        let r, keylength, file_encryption_key =
          match crypt_type with
-         | Pdf.AESV2 -> 4, 128, None
-         | Pdf.AESV3 iso ->
+         | Pdfcryptprimitives.AESV2 -> 4, 128, None
+         | Pdfcryptprimitives.AESV3 iso ->
              begin match oe, ue with
              | Some _, Some ue ->
                  begin match keyfromowner with
@@ -619,7 +634,7 @@ let decrypt_pdf ?keyfromowner user_pw pdf =
                  end
              | _ -> raise (Failure "decrypt_pdf: no oe")
              end
-         | Pdf.ARC4 (k, r) -> r, k, None
+         | Pdfcryptprimitives.ARC4 (k, r) -> r, k, None
        in let encrypt_metadata =
          match Pdf.lookup_direct pdf "/EncryptMetadata" encrypt_dict with
          | Some (Pdf.Boolean false) -> false
@@ -677,9 +692,9 @@ let decrypt_single_stream user_pw owner_pw pdf obj gen stream =
      let crypt_type, u, o, p, id, ue, oe = get_encryption_values pdf in
        let r, keylength =
          match crypt_type with
-         | Pdf.AESV2 -> 4, 128
-         | Pdf.AESV3 b -> (if b then 6 else 5), 256
-         | Pdf.ARC4 (k, r) -> r, k
+         | Pdfcryptprimitives.AESV2 -> 4, 128
+         | Pdfcryptprimitives.AESV3 b -> (if b then 6 else 5), 256
+         | Pdfcryptprimitives.ARC4 (k, r) -> r, k
        in let no_encrypt_metadata =
          match Pdf.lookup_direct pdf "/EncryptMetadata" encrypt_dict with
          | Some (Pdf.Boolean false) -> true
@@ -744,9 +759,9 @@ let key_or_user_password_from_owner ?encryption_values owner_pw pdf =
     in
       let r, keylength =
         match crypt_type with
-        | Pdf.AESV2 -> 4, 128
-        | Pdf.AESV3 x -> (if x then 6 else 5), 256
-        | Pdf.ARC4 (k, r) -> r, k
+        | Pdfcryptprimitives.AESV2 -> 4, 128
+        | Pdfcryptprimitives.AESV3 x -> (if x then 6 else 5), 256
+        | Pdfcryptprimitives.ARC4 (k, r) -> r, k
       in
         if r = 5 || r = 6 then
           if authenticate_owner_password_aesv3 (r = 6) (make_utf8 owner_pw) u o then
@@ -846,7 +861,7 @@ let encrypt_pdf_40bit_inner owner user p user_pw id pdf =
        "/U", Pdf.String user;
        "/P", Pdf.Integer (i32toi p)]
   in
-    match process_cryption false false pdf (Pdf.ARC4 (40, 2)) user_pw 2 user owner p id 40 None with
+    match process_cryption false true pdf (Pdfcryptprimitives.ARC4 (40, 2)) user_pw 2 user owner p id 40 None with
     | Some pdf ->
         {pdf with
           Pdf.trailerdict =
@@ -873,7 +888,7 @@ let encrypt_pdf_128bit_inner owner user p user_pw id pdf =
        "/Length", Pdf.Integer 128;
        "/P", Pdf.Integer (i32toi p)]
   in
-    match process_cryption false false pdf (Pdf.ARC4 (128, 3)) user_pw 3 user owner p id 128 None with
+    match process_cryption false true pdf (Pdfcryptprimitives.ARC4 (128, 3)) user_pw 3 user owner p id 128 None with
     | Some pdf ->
         {pdf with
           Pdf.trailerdict =
@@ -901,7 +916,7 @@ let encrypt_pdf_128bit_inner_r4 owner user p user_pw id pdf encrypt_metadata =
        "/StrF", Pdf.Name "/StdCF";
        "/StmF", Pdf.Name "/StdCF"]
   in
-    match process_cryption (not encrypt_metadata) false pdf (Pdf.ARC4 (128, 4)) user_pw 4 user owner p id 128 None with
+    match process_cryption (not encrypt_metadata) true pdf (Pdfcryptprimitives.ARC4 (128, 4)) user_pw 4 user owner p id 128 None with
     | Some pdf ->
         {pdf with
           Pdf.trailerdict =
@@ -939,7 +954,7 @@ let encrypt_pdf_AES_inner owner user p user_pw id encrypt_metadata pdf =
   in
     match 
       process_cryption
-        (not encrypt_metadata) true pdf Pdf.AESV2 user_pw 4 user owner p id 128 None
+        (not encrypt_metadata) true pdf Pdfcryptprimitives.AESV2 user_pw 4 user owner p id 128 None
     with
     | Some pdf ->
         {pdf with
@@ -981,7 +996,7 @@ let encrypt_pdf_AES256_inner iso encrypt_metadata owner user p perms oe ue id ke
   in
     match
       process_cryption
-        (not encrypt_metadata) true pdf (Pdf.AESV3 iso) "" (if iso then 6 else 5) user owner p id 256 (Some key)
+        (not encrypt_metadata) true pdf (Pdfcryptprimitives.AESV3 iso) "" (if iso then 6 else 5) user owner p id 256 (Some key)
     with
     | Some pdf ->
         {pdf with
@@ -1009,16 +1024,16 @@ let perms_of_p ?digest iso encrypt_metadata p utf8pw o oe u =
 (* Algorithm 3.8. Returns u, ue. *)
 let make_ue iso file_encryption_key user_pw user_validation_salt user_key_salt =
   let hash =
-    if iso then shamix user_pw None else Pdfcryptprimitives.sha256
+    if iso then shamix user_pw None else (fun x -> Pdfcryptprimitives.sha256 (Pdfio.input_of_string x))
   in
   let u =
     String.concat
       ""
-      [(hash (Pdfio.input_of_string (String.concat "" [user_pw; user_validation_salt]))); user_validation_salt; user_key_salt]
+      [(hash (String.concat "" [user_pw; user_validation_salt])); user_validation_salt; user_key_salt]
   in
     let ue = 
       Pdfcryptprimitives.aes_encrypt_data ~firstblock:(int_array_of_string zero_iv) 8
-        (int_array_of_string (hash (Pdfio.input_of_string (String.concat "" [user_pw; user_key_salt]))))
+        (int_array_of_string (hash (String.concat "" [user_pw; user_key_salt])))
         file_encryption_key
     in
       u, String.sub (string_of_bytes ue) 16 32
@@ -1026,15 +1041,15 @@ let make_ue iso file_encryption_key user_pw user_validation_salt user_key_salt =
 (* Algorithm 3.9 *)
 let make_oe iso file_encryption_key owner_pw owner_validation_salt owner_key_salt u =
   let hash =
-    if iso then shamix owner_pw (Some u) else Pdfcryptprimitives.sha256
+    if iso then shamix owner_pw (Some u) else (fun x -> Pdfcryptprimitives.sha256 (Pdfio.input_of_string x))
   in
     let o =
       String.concat
        ""
-       [(hash (Pdfio.input_of_string (String.concat "" [owner_pw; owner_validation_salt; u]))); owner_validation_salt; owner_key_salt]
+       [(hash (String.concat "" [owner_pw; owner_validation_salt; u])); owner_validation_salt; owner_key_salt]
     in
       let digest =
-        int_array_of_string (hash (Pdfio.input_of_string (String.concat "" [owner_pw; owner_key_salt; u])))
+        int_array_of_string (hash (String.concat "" [owner_pw; owner_key_salt; u]))
       in
         let oe = Pdfcryptprimitives.aes_encrypt_data ~firstblock:(int_array_of_string zero_iv) 8 digest file_encryption_key in
           o, String.sub (string_of_bytes oe) 16 32, digest
@@ -1096,7 +1111,7 @@ let recrypt_pdf_user pdf pw =
     | Some x -> (x.Pdf.from_get_encryption_values, x.Pdf.encrypt_metadata, x.Pdf.perms)
   in
     match crypt_type with
-    | Pdf.AESV3 iso ->
+    | Pdfcryptprimitives.AESV3 iso ->
         let oe =
           match oe with
             Some oe -> oe
@@ -1114,13 +1129,13 @@ let recrypt_pdf_user pdf pw =
             encrypt_pdf_AES256_inner
               iso encrypt_metadata o u p perms oe ue id
               (string_of_bytes key) pdf
-    | Pdf.AESV2 ->
+    | Pdfcryptprimitives.AESV2 ->
         encrypt_pdf_AES_inner o u p pw id encrypt_metadata pdf
-    | Pdf.ARC4 (40, _) ->
+    | Pdfcryptprimitives.ARC4 (40, _) ->
         encrypt_pdf_40bit_inner o u p pw id pdf
-    | Pdf.ARC4 (128, 4) ->
+    | Pdfcryptprimitives.ARC4 (128, 4) ->
         encrypt_pdf_128bit_inner_r4 o u p pw id pdf encrypt_metadata
-    | Pdf.ARC4 (128, _) ->
+    | Pdfcryptprimitives.ARC4 (128, _) ->
         encrypt_pdf_128bit_inner o u p pw id pdf
     | _ -> raise (Pdf.PDFError "recrypt_pdf: bad encryption")
 
@@ -1143,7 +1158,7 @@ let recrypt_pdf_owner pdf owner_pw =
       | Some (key, pw) -> (key, pw) 
     in
       match crypt_type with
-      | Pdf.AESV3 iso ->
+      | Pdfcryptprimitives.AESV3 iso ->
           let oe =
             match oe with
               Some oe -> oe
@@ -1155,13 +1170,13 @@ let recrypt_pdf_owner pdf owner_pw =
           in
             encrypt_pdf_AES256_inner
               iso encrypt_metadata o u p perms oe ue id key pdf
-      | Pdf.AESV2 ->
+      | Pdfcryptprimitives.AESV2 ->
           encrypt_pdf_AES_inner o u p pw id encrypt_metadata pdf
-      | Pdf.ARC4 (40, _) ->
+      | Pdfcryptprimitives.ARC4 (40, _) ->
           encrypt_pdf_40bit_inner o u p pw id pdf
-      | Pdf.ARC4 (128, 4) ->
+      | Pdfcryptprimitives.ARC4 (128, 4) ->
           encrypt_pdf_128bit_inner_r4 o u p pw id pdf encrypt_metadata
-      | Pdf.ARC4 (128, _) ->
+      | Pdfcryptprimitives.ARC4 (128, _) ->
           encrypt_pdf_128bit_inner o u p pw id pdf
       | _ -> raise (Pdf.PDFError "recrypt_pdf_owner: bad encryption")
 
