@@ -31,3 +31,17 @@ val sha384 : Pdfio.input -> string
 (** SHA512 digest *)
 val sha512 : Pdfio.input -> string
 
+(**/**)
+type encryption = 
+  | ARC4 of int * int
+  | AESV2
+  | AESV3 of bool (* true = iso, false = old algorithm *)
+
+val find_hash : encryption -> int32 -> int32 -> int array -> int -> int array
+
+val decrypt_stream_data :
+  encryption ->
+  bool ->
+  string option ->
+  int -> int -> int array -> int -> int -> Pdfio.bytes -> Pdfio.bytes
+
