@@ -227,18 +227,11 @@ let input_of_stream ?(source = "bytes") s =
   let offset = ref 0 in
     let ssize = bytes_size !(s.data) in
       let input_int () =
-        if s.pos > ssize - 1
-          then
-            begin
-              s.pos <- s.pos + 1;
-              no_more
-            end
-          else
-            begin
-              let r = bget_unsafe !(s.data) s.pos in
-                s.pos <- s.pos + 1;
-                r
-            end
+        let r =
+          if s.pos > ssize - 1 then no_more else bget_unsafe !(s.data) s.pos
+        in
+          s.pos <- s.pos + 1;
+          r
       in
         {pos_in =
            (fun () -> s.pos - !offset);
@@ -275,7 +268,7 @@ let input_of_string ?(source = "string") s =
        else
          begin
            pos := !pos + 1;
-           Some (s.[!pos - 1])
+           Some (String.unsafe_get s (!pos - 1))
          end
    in
      {pos_in =
@@ -301,9 +294,11 @@ let input_of_string ?(source = "string") s =
       in_channel_length =
         String.length s;
       set_offset =
-        (fun _ -> raise (Failure "input_of_string: no set_offset"));
-      caml_channel = None;
-      source = source}
+        (fun _ -> ());
+      caml_channel =
+        None;
+      source =
+        source}
 
 (* Output functions over channels *)
 let output_of_channel ch =
@@ -457,13 +452,22 @@ let setinit i s o l =
 let setinit_string i s o l =
   let max = String.length s - 1
   and last = o + 1 - 1 in
-    if o > max || o < 0 || last < 0 || last > max then raise (Failure "setinit") else
+    if o > max || o < 0 || last < 0 || last > max then raise (Failure "setinit_string") else
       match i.caml_channel with
       | Some ch ->
           really_input ch s o l
       | None ->
           for x = o to o + l - 1 do String.unsafe_set s x (Char.unsafe_chr (i.input_byte ())) done
 
+let bytes_of_input i o l =
+  i.seek_in o;
+  let s = String.create l in
+    setinit_string i s 0 l;
+    if l <= Sys.max_string_length then
+      Short s
+    else
+      bytes_of_string s
+
 let getinit i s o l =
   let max = bytes_size s - 1
   and last = o + 1 - 1 in
