@@ -114,9 +114,11 @@ let debug_pagelabels ls =
 let debug_collection_of_pagelabels =
   iter (fun ls -> debug_pagelabels ls; flprint "\n")
 
-(* Merging /Dests (Named destinations). Since the new /Dests must be an indirect
-reference, we add the new object to the pdf, returning the new pdf and the
-reference. *)
+(* Merging /Dests (Named destinations) in the catalog (PDF v1.1 style, rather
+ * than in the PDF 1.3 style in the name tree). Since the new /Dests must be an
+ * indirect reference, we add the new object to the pdf, returning the new pdf
+ * and the reference.
+FIXME: merging a v1.1 file with a v1.2 file will result in both sets of dests, confusing the reader...*)
 let new_dests pdf pdfs =
   let dests =
     option_map
@@ -127,10 +129,11 @@ let new_dests pdf pdfs =
           | _ -> None)
       pdfs
   in
-    let new_dests =
-      Pdf.Dictionary (flatten dests)
-    in
-      Pdf.addobj pdf new_dests
+    if dests = [] then None else
+      let new_dests =
+        Pdf.Dictionary (flatten dests)
+      in
+        Some (Pdf.addobj pdf new_dests)
 
 (* Names distinguish PDFs which are actually the same. So we only use the first
 of each group of same ones. Then renumber them. and return. *)
@@ -148,11 +151,6 @@ let merge_pdfs_renumber names pdfs =
         let table = combine first_names (Pdf.renumber_pdfs first_pdfs) in
           map (function k -> lookup_failnull k table) names
       
-(*
-  let first_names, first_pdfs = split (setify (combine names pdfs)) in
-    let table = combine first_names (Pdf.renumber_pdfs first_pdfs) in
-      map (function k -> lookup_failnull k table) names*)
-
 (* Reading a name tree, flattened. *)
 let rec read_name_tree pdf tree =
   let names =
@@ -225,33 +223,22 @@ let merge_namedicts pdf pdfs =
             (* Return the new pdf, and the new dictionary. *)
             Pdf.addobj pdf newdict
 
-(* Things which we don't yet properly merge, are just copied across from the
-first document, if they're there. *)
-let copied_from_first_document pdf =
-  let names =
-    ["/Threads";
-     "/AcroForm";
-     "/AA";
-     "/URI";
-     "/StructTreeRoot";
-     "/SpiderInfo";
-     "/OCProperties"]
-  in let catalog = Pdf.catalog_of_pdf pdf in
-    option_map
-      (function name ->
-        match Pdf.lookup_direct pdf name catalog with
-        | None -> None
-        | Some x -> Some (name, x))
-      names
+(* Merge catalog items from the PDFs, taking an abitrary instance of any one we
+ * find. Items we know how to merge properly, like /Dests, /Names, /PageLabels,
+ * /Outlines, will be overwritten, so we don't worry about them here. *)
+let catalog_items_from_original_documents pdfs =
+  let catalog_entries =
+    flatten
+      (map
+        (fun pdf ->
+           match Pdf.catalog_of_pdf pdf with
+             Pdf.Dictionary es -> es
+           | _ -> failwith "catalog_items_from_original_documents")
+        pdfs)
+  in
+    fold_left (fun d (k, v) -> add k v d) [] catalog_entries
 
-let merge_pdfs retain_numbering do_remove_duplicate_fonts (names : string list) pdfs ranges =
-  (*Printf.printf "merge_pdfs: retain_numbering = %b, do_remove_duplicate_fonts = %b \n %i names, %i pdfs, %i ranges\n"
-  retain_numbering do_remove_duplicate_fonts (length names) (length pdfs) (length ranges);
-  flprint "names:\n";
-  iter (Printf.printf "%s\n") names;
-  (*flprint "ranges:\n";
-  iter (Printf.printf "%s\n") (map string_of_range ranges);*)
-  flprint "ONE";*)
+let merge_pdfs retain_numbering do_remove_duplicate_fonts names pdfs ranges =
   let pdfs = merge_pdfs_renumber names pdfs in
     let minor' = fold_left max 0 (map (fun p -> p.Pdf.minor) pdfs) in
       let pagelists = map Pdfpage.pages_of_pagetree pdfs
@@ -261,41 +248,45 @@ let merge_pdfs retain_numbering do_remove_duplicate_fonts (names : string list)
             iter (fun n -> pages =| select n pagelist) range;
             rev !pages
         in
-          let pages = flatten (map2 select_pages ranges pagelists) in
-            iter (Pdf.objiter (fun k v -> ignore (Pdf.addobj_given_num pdf (k, v)))) pdfs;
-            (* Make the hints for preserving... *)
-            let pdf, pagetree_num = Pdfpage.add_pagetree pages pdf in
-              let page_labels =
-                if retain_numbering
-                 then Pdfpagelabels.merge_pagelabels pdfs ranges
-                 else []
+  let pages = flatten (map2 select_pages ranges pagelists) in
+    iter (Pdf.objiter (fun k v -> ignore (Pdf.addobj_given_num pdf (k, v)))) pdfs;
+    let pdf, pagetree_num = Pdfpage.add_pagetree pages pdf in
+      let page_labels =
+        if retain_numbering
+         then Pdfpagelabels.merge_pagelabels pdfs ranges
+         else []
+      in
+        let dests = new_dests pdf pdfs in
+          let namedict = merge_namedicts pdf pdfs in
+            let extra_catalog_entries =
+              let with_names =
+                (add  "/Names" (Pdf.Indirect namedict)
+                  (catalog_items_from_original_documents pdfs))
               in
-                let dests = new_dests pdf pdfs in
-                  let namedict = merge_namedicts pdf pdfs in
-                    let extra_catalog_entries =
-                      [("/Dests", Pdf.Indirect dests);
-                       ("/Names", Pdf.Indirect namedict)]
-                      @ copied_from_first_document (hd pdfs)
-                    in
-                      let pdf = Pdfpage.add_root pagetree_num extra_catalog_entries pdf in
-                      (* To sort out annotations etc. *)
-                      let old_page_numbers =
-                        let select_page_numbers range pageobjnums =
-                          let pages = ref [] in
-                            iter (fun n -> pages =| select n pageobjnums) range;
-                            rev !pages
-                        in
-                          flatten (map2 select_page_numbers ranges (map Pdf.page_reference_numbers pdfs))
-                      in let new_page_numbers =
-                        Pdf.page_reference_numbers pdf
-                      in
-                        let changes = combine old_page_numbers new_page_numbers in
-                          Pdf.objselfmap
-                          (Pdf.renumber_object_parsed pdf (hashtable_of_dictionary changes))
-                          pdf;
-                        let pdf = {pdf with Pdf.major = 1; Pdf.minor = minor'} in
-                          let pdf = merge_bookmarks changes pdfs ranges pdf in
-                            Pdfpagelabels.write pdf page_labels;
-                            if do_remove_duplicate_fonts then remove_duplicate_fonts pdf;
-                            pdf
+                match dests with
+                  None -> with_names
+                | Some dests ->
+                    add "/Dests" (Pdf.Indirect dests) with_names
+            in
+   let pdf = Pdfpage.add_root pagetree_num extra_catalog_entries pdf in
+      (* To sort out annotations etc. *)
+      let old_page_numbers =
+        let select_page_numbers range pageobjnums =
+          let pages = ref [] in
+            iter (fun n -> pages =| select n pageobjnums) range;
+            rev !pages
+        in
+          flatten (map2 select_page_numbers ranges (map Pdf.page_reference_numbers pdfs))
+      in let new_page_numbers =
+        Pdf.page_reference_numbers pdf
+      in
+        let changes = combine old_page_numbers new_page_numbers in
+          Pdf.objselfmap
+          (Pdf.renumber_object_parsed pdf (hashtable_of_dictionary changes))
+          pdf;
+   let pdf = {pdf with Pdf.major = 1; Pdf.minor = minor'} in
+     let pdf = merge_bookmarks changes pdfs ranges pdf in
+       Pdfpagelabels.write pdf page_labels;
+       if do_remove_duplicate_fonts then remove_duplicate_fonts pdf;
+       pdf
 
