@@ -13,7 +13,7 @@ containing the graphical content stream (see the Pdfops module), mediabox
 the page size, resources the page's resource dictionary, rotate its rotation
 and rest any other entries to reside in the page dictionary. *)
 type t =
-  {content : Pdf.pdfobject list;
+  {content : Pdf.pdfobject list; (*FIXME Change this to int list, to ensure sharing? *)
    mediabox : Pdf.pdfobject;
    resources : Pdf.pdfobject;
    rotate : rotation;
@@ -114,7 +114,25 @@ let rec find_pages pages pdf resources mediabox rotate =
         | Some x -> Some x
         | None -> mediabox
       in let contents =
-        Pdf.lookup_direct pdf "/Contents" pages
+        (* 28th March 2016. We modify this to always create an array of one
+        indirect if just a single contents stream, rather than following
+        through to the actual object. Other code can then preserve the
+        sharing. *)
+        begin match pages with
+          Pdf.Dictionary d ->
+            begin match lookup "/Contents" d with
+              Some (Pdf.Indirect i) ->
+                (* A single content stream, or indirect to array *)
+                begin match Pdf.lookup_obj pdf i with
+                | Pdf.Array a -> Some (Pdf.Array a)
+                | _ -> Some (Pdf.Array [Pdf.Indirect i])
+                end
+            | _ ->
+                (* An array of indirects. Just return it *)
+                Pdf.lookup_direct pdf "/Contents" pages
+            end
+        | _ -> assert false
+        end
       in let rotate =
         match Pdf.lookup_direct pdf "/Rotate" pages with
         | Some (Pdf.Integer r) -> rotation_of_int r
@@ -135,12 +153,13 @@ let rec find_pages pages pdf resources mediabox rotate =
                   (function x ->
                      match Pdf.direct pdf x with
                      | Pdf.Stream _ -> x
-                     | _ -> raise (Pdf.PDFError "Bad /Contents"))
+                     | x ->
+                         raise (Pdf.PDFError ("Bad /Contents 1")))
                   cs;
             | Some pdfobject ->
                 begin match Pdf.direct pdf pdfobject with
                 | Pdf.Stream _ -> [pdfobject]
-                | _ -> raise (Pdf.PDFError "Bad /Contents")
+                | _ -> raise (Pdf.PDFError "Bad /Contents 2")
                 end);
           mediabox =
             (match mediabox with
@@ -260,7 +279,13 @@ let change_operator pdf lookup lookup_option seqnum = function
   | Pdfops.Op_DP (n, Pdf.Name p) ->
       Pdfops.Op_DP (n, Pdf.Name (lookup "/Properties" seqnum p))
   | Pdfops.Op_BDC (n, Pdf.Name p) ->
-      Pdfops.Op_BDC (n, Pdf.Name (lookup "/Properties" seqnum p))
+      begin match lookup_option "/Properties" seqnum p with
+        | Some x ->
+            Pdfops.Op_BDC (n, Pdf.Name x)
+        | None ->
+            Printf.eprintf "Warning: Missing Op_BDC /Properties entry\n";
+            Pdfops.Op_BDC (n, Pdf.Name p)
+      end
   | Pdfops.InlineImage (dict, bytes) ->
       (* Replace any indirect "/CS" or "/ColorSpace" with a new "/CS" *)
       let dict' =
@@ -283,6 +308,8 @@ let change_operator pdf lookup lookup_option seqnum = function
         Pdfops.InlineImage (dict', bytes)
   | x -> x
 
+(* Only for use with twoup now. FIXME: Can blow up shared content streams. Needs
+a cunning new method to preserve sharing. *)
 let renumber_pages pdf pages =
   match pages with
   | [] -> []
@@ -525,6 +552,11 @@ let add_pagetree pages pdf =
         let ptree = pagetree getobjnum pages 0 in
           let objects = objects_of_ptree getobjnum extras ptree in
             let topnode = match hd objects with (n, _) -> n in
+              (*Printf.printf "There were %i objects_of_ptree\n" (List.length objects);
+              List.iter (fun (i, x) -> Printf.printf "%i: %s\n" i (Pdfwrite.string_of_pdf x)) objects;
+              Printf.printf "There were %i extras\n" (List.length !extras);
+              List.iter (fun (i, x) -> Printf.printf "%i: %s\n" i
+              (Pdfwrite.string_of_pdf x)) !extras;*)
               iter (fun x -> ignore (Pdf.addobj_given_num pdf x)) (objects @ !extras);
               pdf, topnode
 
@@ -560,8 +592,11 @@ let add_root pageroot extras pdf =
 
 Other objects (e.g destinations in the document outline) may point to the
 individual page objects, so we must renumber these. We can only do this if the
-number of pages are the same. We do this [if replace_numbers is true]. *)
-let change_pages change_references basepdf pages' =
+number of pages are the same. We do this [if change_references is true]. If the
+new and old page lists are of differenct lengths, change_references must be
+false, or you must supply the changes (expressed as (from, to) 1-based serial
+number pairs) *)
+let change_pages ?changes change_references basepdf pages' =
   let pdf = Pdf.empty () in
     Pdf.objiter (fun k v -> ignore (Pdf.addobj_given_num pdf (k, v))) basepdf;
     let old_page_numbers = Pdf.page_reference_numbers basepdf in
@@ -583,14 +618,30 @@ let change_pages change_references basepdf pages' =
         in
           let pdf = add_root pagetree_num existing_root_entries pdf in
             let new_page_numbers = Pdf.page_reference_numbers pdf in
-              if change_references && length old_page_numbers = length new_page_numbers
-                then
-                  let changes = combine old_page_numbers new_page_numbers in
-                    Pdf.objselfmap
-                      (Pdf.renumber_object_parsed pdf (hashtable_of_dictionary changes))
-                      pdf;
-                    pdf
-                else
+              if not change_references then pdf else
+                let changes =
+                   match changes with
+                     None ->
+                       if length old_page_numbers = length new_page_numbers then
+                         combine old_page_numbers new_page_numbers
+                       else
+                         begin
+                           Printf.printf "change_pages: No change supplied, and lengths differ\n";
+                           []
+                         end
+                   | Some cs ->
+                       (* Turn the 1-based serial numbers into page reference numbers *)
+                       try
+                         List.map
+                           (fun (x, y) ->
+                            List.nth old_page_numbers (x - 1), List.nth new_page_numbers (y - 1))
+                         cs
+                       with
+                         _ -> raise (Pdf.PDFError "change_pages: bad serial number")
+                in
+                  Pdf.objselfmap
+                    (Pdf.renumber_object_parsed pdf (hashtable_of_dictionary changes))
+                    pdf;
                   pdf
 
 (* Return a pdf with a subset of pages, but nothing else changed - exactly the
@@ -618,7 +669,7 @@ let page_object_number pdf destpage =
     Some (select destpage (Pdf.page_reference_numbers pdf))
   with
     (* The page might not exist in the output *)
-    Invalid_argument "select" -> None
+    Invalid_argument _ (*"select"*) -> None
 
 let target_of_pagenumber pdf i =
   match page_object_number pdf i with
@@ -738,9 +789,10 @@ let rec fixup_duplicate_pages pdf =
 
 let pdf_of_pages ?(retain_numbering = false) basepdf range =
   let page_labels =
-    if retain_numbering
-      then Pdfpagelabels.merge_pagelabels [basepdf] [range]
-      else []
+    if List.length (Pdfpagelabels.read basepdf) = 0 then [] else
+      if retain_numbering
+        then Pdfpagelabels.merge_pagelabels [basepdf] [range]
+        else []
   and marks =
     let refnums = Pdf.page_reference_numbers basepdf in
       option_map
@@ -889,14 +941,178 @@ let postpend_operators pdf ops ?(fast=false) page =
       {page with content =
          [Pdfops.stream_of_ops (beforeops @ Pdfops.parse_operators pdf page.resources page.content @ afterops)]}
 
-(* (* Ensure that there are no inherited attributes in the page tree --- in other
-words they are all explicit. This is required before writing a file with
-linearization *)
-let pagetree_make_explicit pdf =
-  let pages = pages_of_pagetree pdf in
-    change_pages true pdf pages*)
+(* Source of possible prefix strings. String is always copied. *)
+let next_string s =
+  if s = "" then "a" else
+    if s.[0] = 'z' then "a" ^ s else
+      let s' = String.copy s in
+         s'.[0] <- char_of_int (int_of_char s'.[0] + 1);
+         s'
 
-(* Set the reference in Pdfwrite for inter-module recursion. *)
-(*let _ =
-  Pdfwrite.pagetree_make_explicit := pagetree_make_explicit*)
+(* True if one string [p] is a prefix of another [n] *)
+let is_prefix p n =
+  String.length p <= String.length n &&
+  String.sub n 0 (String.length p) = p
+
+(* a) List every name used in a /Resources in a /Type /Page or
+ /Type /Pages (without the leading "/"
+   b) Find the shortest lower-case alphabetic string which is not a prefix of any of these
+   strings. This prefix can be added to the other PDF's names, and will never
+   clash with any of these. *)
+let names_used pdf =
+  let names = ref [] in
+  let unslash x =
+    if x = "" then "" else String.sub x 0 (String.length x - 1)
+  in
+    Pdf.objiter
+      (fun n obj ->
+        match obj with
+          Pdf.Dictionary d | Pdf.Stream {contents = (Pdf.Dictionary d, _)} ->
+            begin match lookup "/Type" d with
+              Some (Pdf.Name ("/Page" | "/Pages")) ->
+                begin match Pdf.lookup_direct pdf "/Resources" obj with
+                  Some resources ->
+                    List.iter
+                      (fun key ->
+                         match Pdf.lookup_direct pdf key resources with
+                           Some (Pdf.Dictionary d) ->
+                             List.iter
+                               (fun (k, _) -> names := unslash k::!names)
+                               d
+                         | _ -> ())
+                      resource_keys
+                | _ -> ()
+                end
+            | _ -> ()
+            end
+        | _ -> ()
+      )
+      pdf;
+    setify !names
+
+let shortest names =
+  let rec loop prefix =
+    if List.exists (is_prefix prefix) names
+      then loop (next_string prefix)
+      else prefix
+  in
+    loop "a"
+
+let shortest_unused_prefix pdf =
+  shortest (names_used pdf)
+
+let addp p n =
+  if n = "" then raise (Pdf.PDFError "addp: blank name") else
+    "/" ^ p ^ String.sub n 1 (String.length n - 1)
+
+let prefix_operator pdf p = function
+  | Pdfops.Op_Tf (f, s) -> Pdfops.Op_Tf (addp p f, s)
+  | Pdfops.Op_gs n -> Pdfops.Op_gs (addp p n)
+  | Pdfops.Op_CS n -> Pdfops.Op_CS (addp p n)
+  | Pdfops.Op_cs n -> Pdfops.Op_cs (addp p n)
+  | Pdfops.Op_SCNName (s, ns) -> Pdfops.Op_SCNName (addp p s, ns)
+  | Pdfops.Op_scnName (s, ns) -> Pdfops.Op_scnName (addp p s, ns)
+  | Pdfops.Op_sh s -> Pdfops.Op_sh (addp p s)
+  | Pdfops.Op_Do x -> Pdfops.Op_Do (addp p x)
+  | Pdfops.Op_DP (n, Pdf.Name x) -> Pdfops.Op_DP (n, Pdf.Name (addp p x))
+  | Pdfops.Op_BDC (n, Pdf.Name x) -> Pdfops.Op_BDC (n, Pdf.Name (addp p x))
+  | Pdfops.InlineImage (dict, bytes) ->
+      (* Replace any indirect "/CS" or "/ColorSpace" with a new "/CS" *)
+      let dict' =
+        match Pdf.lookup_direct_orelse pdf "/CS" "/ColorSpace" dict with
+        | Some (Pdf.Name "/DeviceGray")
+        | Some (Pdf.Name "/DeviceRGB")
+        | Some (Pdf.Name "/DeviceCMYK")
+        | Some (Pdf.Name "/G")
+        | Some (Pdf.Name "/RGB")
+        | Some (Pdf.Name "/CMYK") -> dict
+        | Some (Pdf.Name n) ->
+            Pdf.add_dict_entry
+              (Pdf.remove_dict_entry
+                (Pdf.remove_dict_entry dict "/ColorSpace")
+                "/CS")
+              "/CS"
+              (Pdf.Name (addp p n))
+        | _ -> dict
+      in
+        Pdfops.InlineImage (dict', bytes)
+  | x -> x
+
+let change_resources pdf prefix resources =
+  let newdict name =
+    match Pdf.lookup_direct pdf name resources with
+    | Some (Pdf.Dictionary fonts) ->
+        Pdf.Dictionary (map (fun (k, v) -> addp prefix k, v) fonts)
+    | _ -> Pdf.Dictionary []
+  in
+    let newdicts = map newdict resource_keys in
+      let resources = ref resources in
+        iter2
+          (fun k v ->
+            resources := Pdf.add_dict_entry !resources k v)
+          resource_keys
+          newdicts;
+        !resources
+
+(* For each object in the PDF with /Type /Page or /Type /Pages:
+  a) Add the prefix to any name in /Resources
+  b) Add the prefix to any name used in any content streams, keeping track of
+  the streams we have processed to preserve sharing
+  
+FIXME: If a non-ISO PDF with content streams which don't end on lexical boundaries
+is provided, the parse will fail, and this function will raise an exception.
+
+The long-term solution to this is to explode the PDF with an
+-unshare-content-streams option, and then fix squeeze to re-share subcontent
+streams *)
+let add_prefix pdf prefix =
+  let fixed_streams = Hashtbl.create 100 in
+  let fix_stream resources i =
+    match i with Pdf.Indirect i ->
+      if not (Hashtbl.mem fixed_streams i) then
+        let operators = Pdfops.parse_operators pdf resources [Pdf.Indirect i] in
+          let operators' = map (prefix_operator pdf prefix) operators in
+            begin match Pdf.lookup_obj pdf i with
+              Pdf.Stream ({contents = (dict, stream)} as s) ->
+                begin match Pdfops.stream_of_ops operators' with
+                  Pdf.Stream {contents = ncontents} -> s := ncontents
+                | _ -> failwith "add_prefix: bad stream"
+                end
+            | _ -> failwith "add_prefix: bad stream 2"
+            end;
+            Hashtbl.add fixed_streams i ()
+    | _ -> failwith "add_prefix: not indirect"
+  in
+  Pdf.objselfmap
+    (fun obj ->
+       match obj with
+         Pdf.Dictionary dict as d ->
+           begin match Pdf.lookup_direct pdf "/Type" d with
+             Some (Pdf.Name ("/Page" | "/Pages")) ->
+               let resources =
+                 begin match Pdf.lookup_direct pdf "/Resources" obj with
+                   Some resources -> Some (change_resources pdf prefix resources)
+                 | _ -> (); None
+                 end
+               in
+                 begin match lookup "/Contents" dict with
+                   Some (Pdf.Indirect i) ->
+                     fix_stream
+                       (if resources = None then Pdf.Dictionary [] else unopt resources)
+                       (Pdf.Indirect i)
+                 | Some (Pdf.Array a) ->
+                     List.iter
+                       (fix_stream
+                         (if resources = None then Pdf.Dictionary [] else unopt resources))
+                       a
+                 | _ -> ()
+                 end;
+                 begin match resources with
+                   Some resources -> Pdf.add_dict_entry d "/Resources" resources
+                 | None -> d
+                 end
+           | _ -> obj
+           end
+       | _ -> obj)
+    pdf
 
