@@ -63,8 +63,9 @@ val renumber_pages : Pdf.t -> t list -> t list
 (** Change the pages in a document for some new ones. If the boolean is true
 and the number of pages in the old and new documents are equal, references to
 the old pages from outside the page tree (for instance in destinations or
-bookmarks) are renumbered. This ensures bookmarks are preserved correctly. *)
-val change_pages : bool -> Pdf.t -> t list -> Pdf.t
+bookmarks) are renumbered. This ensures bookmarks are preserved correctly.
+[is_combine_pages] is a special undocumented mode, false by default *)
+val change_pages : ?changes:((int * int) list) -> bool -> Pdf.t -> t list -> Pdf.t
 
 (** Return a pdf with a subset of pages, but nothing else changed - exactly the
 same page object numbers, so bookmarks etc still work. Also sorts out bookmarks
@@ -73,6 +74,10 @@ val pdf_of_pages : ?retain_numbering:bool -> Pdf.t -> int list -> Pdf.t
 
 (**/**)
 
+val shortest_unused_prefix : Pdf.t -> string
+
+val add_prefix : Pdf.t -> string -> unit
+
 val prepend_operators : Pdf.t -> Pdfops.t list -> ?fast:bool -> t -> t
 
 val postpend_operators : Pdf.t -> Pdfops.t list -> ?fast:bool -> t -> t
