@@ -44,9 +44,8 @@ let backline i =
   read_back_until is_newline i;
   nudge i
 
-(* Read the major and minor version numbers  from a PDF [1.x] file. Fail if
-header invalid or major version number is not 1.  *)
-let get9chars i =
+(* Read the major and minor version numbers *)
+let get8chars i =
   let c1 = i.input_char () in
   let c2 = i.input_char () in
   let c3 = i.input_char () in
@@ -55,15 +54,14 @@ let get9chars i =
   let c6 = i.input_char () in
   let c7 = i.input_char () in
   let c8 = i.input_char () in
-  let c9 = i.input_char () in
-    try map unopt [c1; c2; c3; c4; c5; c6; c7; c8; c9] with _ -> []
+    try map unopt [c1; c2; c3; c4; c5; c6; c7; c8] with _ -> []
 
 let rec read_header_inner pos i =
   try
     if pos > 1024 then raise End_of_file else
-      i.seek_in pos;
-      match get9chars i with
-      | '%'::'P'::'D'::'F'::'-'::_::'.'::minor ->
+      i.Pdfio.seek_in pos;
+      match get8chars i with
+      | '%'::'P'::'D'::'F'::'-'::major::'.'::minor ->
           let minorchars = takewhile isdigit minor in
             if minorchars = []
               then
@@ -72,12 +70,12 @@ let rec read_header_inner pos i =
               else
                 begin
                   i.set_offset pos;
-                  1, int_of_string (implode minorchars)
+                  int_of_string (string_of_char major), int_of_string (implode minorchars)
                 end
       | _ ->
           read_header_inner (pos + 1) i
   with
-    End_of_file | Failure "int_of_string" ->
+    End_of_file | Failure _ (*"int_of_string"*) ->
       raise (Pdf.PDFError (Pdf.input_pdferror i "Could not read PDF header"))
 
 let read_header =
@@ -210,9 +208,10 @@ let lex_number i =
       | Pdfgenlex.LexReal f -> LexReal f
       | _ -> LexNone
     with
-    | Failure "hd" -> LexNone
+    (* FIXME: Heed warning 52 *)
+    | Failure x when x = "hd" -> LexNone
     | Pdf.PDFError _ (* can't cope with floats with leading point. *)
-    | Failure "int_of_string" ->
+    | Failure _ (*"int_of_string"*) ->
         LexReal
           (float_of_string
              (i.seek_in pos; (getuntil_white_or_delimiter_string i)))
@@ -327,7 +326,7 @@ let lex_string i =
     done;
     LexString (Buffer.contents str)
   with
-    | Failure "unopt" ->
+    | Failure _ (*"unopt"*) ->
         raise (Pdf.PDFError (Pdf.input_pdferror i "lex_string failure"))
 
 (* Lex a hexadecimal string. *)
@@ -360,7 +359,7 @@ let lex_hexstring i =
               LexString (Buffer.contents str)
        with
          | End_of_file -> LexString (Buffer.contents str)
-         | Failure "unopt" ->
+         | Failure _ (*"unopt"*) ->
              raise (Pdf.PDFError (Pdf.input_pdferror i "lex_hexstring"))
 
 (* Lex a keyword. *)
@@ -480,7 +479,7 @@ let lex_stream_data i l opt =
           i.seek_in (pos + l);
           if is_malformed i
             then (i.seek_in original_pos; lex_malformed_stream_data i)
-            else (i.seek_in (pos + l); LexStream (Pdf.ToGet (i, pos, l)))
+            else (i.seek_in (pos + l); LexStream (Pdf.ToGet (Pdf.toget i pos l)))
         end
     with
       _ -> raise (Pdf.PDFError (Pdf.input_pdferror i "lex_stream_data"))
@@ -834,9 +833,7 @@ let rec parse_to_tree (sofar : partial_parse_element list) = function
 let parse_finish ?(failure_is_ok = false) q =
   match q with
   | [Parsed (Pdf.Integer o); Parsed (Pdf.Integer _);
-    Lexeme LexObj; Parsed obj; Lexeme LexEndObj]
-  | [Parsed (Pdf.Integer o); Parsed (Pdf.Integer _);
-    Lexeme LexObj; Parsed obj] ->
+    Lexeme LexObj; Parsed obj; Lexeme LexEndObj] ->
       o, obj
   | Parsed (Pdf.Integer o)::
     Parsed (Pdf.Integer _)::
@@ -846,7 +843,7 @@ let parse_finish ?(failure_is_ok = false) q =
     Lexeme LexEndStream::_ ->
       (* Fix up length, if necessary *)
       let l =
-        match s with Pdf.Got b -> bytes_size b | Pdf.ToGet (_, _, l) -> l
+        match s with Pdf.Got b -> bytes_size b | Pdf.ToGet t -> Pdf.length_of_toget t
       and lold =
         try
           begin match
@@ -862,13 +859,22 @@ let parse_finish ?(failure_is_ok = false) q =
                {contents = Pdf.add_dict_entry obj "/Length" (Pdf.Integer l), s})
           else
             (o, Pdf.Stream {contents = obj, s})
+  | Parsed (Pdf.Integer o)::Parsed (Pdf.Integer _)::
+    Lexeme LexObj::Parsed obj::_ ->
+      o, obj
   | [Parsed d] ->
       0, d
   | [Parsed (Pdf.Integer o); Parsed (Pdf.Integer _);
      Lexeme LexObj; Lexeme LexEndObj] ->
       o, Pdf.Null
   | l ->
-      raise (Pdf.PDFError "Could not extract object")
+      Printf.eprintf "list length %i\n" (length l);
+        List.iter
+          (function
+              Parsed p -> Printf.eprintf "%s\n" (Pdfwrite.string_of_pdf p)
+            | Lexeme l -> Printf.eprintf "%s\n" (string_of_lexeme l))
+          l;
+        raise (Pdf.PDFError "Could not extract object")
 
 (* Parse some lexemes *)
 let parse ?(failure_is_ok = false) lexemes =
@@ -876,7 +882,9 @@ let parse ?(failure_is_ok = false) lexemes =
     parse_finish
       ~failure_is_ok:failure_is_ok (parse_to_tree [] (parse_R lexemes))
   with
-    Pdf.PDFError _ -> (max_int, Pdf.Null)
+    (* 14th November 2016: fixed this up to re-raise. Check *)
+    Pdf.PDFError _ as e ->
+      if failure_is_ok then (max_int, Pdf.Null) else raise e
 
 (* Given an object stream pdfobject and a list of object indexes to extract,
  return an [(int * Pdf.objectdata) list] representing those object number,
@@ -886,9 +894,9 @@ let lex_stream_object
 =
   if !read_debug then
     begin
-      Printf.printf "lexing object stream at %i\n" (i.Pdfio.pos_in ());
-      Printf.printf "lexing object stream %i\nTo find the indexes:\n" obj;
-      iter (Printf.printf "%i ") indexes; flprint "\n"
+      Printf.eprintf "lexing object stream at %i\n" (i.Pdfio.pos_in ());
+      Printf.eprintf "lexing object stream %i\nTo find the indexes:\n" obj;
+      iter (Printf.eprintf "%i ") indexes; Printf.eprintf "\n"
     end;
   let _, stmobj = parse (lex_object i xrefs parse opt obj) in
     match stmobj with
@@ -1068,7 +1076,8 @@ let read_xref i =
             | _ -> () (* Xref stream types won't have been generated. *)
           done
         with
-          End_of_file | Sys_error _ | Failure "int_of_string"-> fail ()
+          End_of_file | Sys_error _
+          | Failure _ (*"int_of_string"*) -> fail ()
       end;
       !xrefs
 
@@ -1091,7 +1100,7 @@ let read_xref_line_stream i w1 w2 w3 =
       let f1 = read_field w1 in
         let f2 = read_field w2 in
           let f3 = read_field w3 in
-            if !read_debug then Printf.printf "%i %i %i\n" f1 f2 f3;
+            (*if !read_debug then Printf.eprintf "%i %i %i\n" f1 f2 f3;*)
             match f1 with
             | 0 -> StreamFree (f2, f3)
             | 1 -> Valid (f2, f3)
@@ -1116,7 +1125,7 @@ let read_xref_stream i =
       | _ -> Printf.eprintf "couldn't lex object number\n"; raise err
     in
     if !read_debug then
-      Printf.printf
+      Printf.eprintf
         "Object number of this xref stream is %i\n" xrefstream_objectnumber;
     (* Go back to the beginning *)
     i.seek_in original_pos;
@@ -1159,7 +1168,7 @@ let read_xref_stream i =
         | _ -> raise err
       in
         if !read_debug then
-          flprint "HAVE READ XREF STREAM DICT, NOW ACTUAL XREF STREAM DATA\n";
+          Printf.eprintf "HAVE READ XREF STREAM DICT, NOW ACTUAL XREF STREAM DATA\n";
         Pdfcodec.decode_pdfstream (Pdf.empty ()) stream;
         let w1, w2, w3 =
           match Pdf.lookup_direct (Pdf.empty ()) "/W" stream with
@@ -1172,14 +1181,20 @@ let read_xref_stream i =
           | _ -> raise err
         in let xrefs = ref [] in
           begin try
+            if !read_debug then
+              Printf.printf "About to start read_xref_stream\n%!";
             while true do xrefs =| read_xref_line_stream i' w1 w2 w3 done
+
           with
-            _ -> ()
+            _ ->
+              if !read_debug then
+                Printf.printf "End of read_xref_stream\n%!";
+              ()
           end;
           xrefs := rev !xrefs;
           if !read_debug then
-            Printf.printf
-              "****** read %i raw Xref stream entries\n" (length !xrefs);
+            Printf.eprintf
+              "****** read %i raw Xref stream entries\n%!" (length !xrefs);
           let starts_and_lens =
             match Pdf.lookup_direct (Pdf.empty ()) "/Index" stream with
             | Some (Pdf.Array elts) ->
@@ -1207,6 +1222,9 @@ let read_xref_stream i =
                   [0, size]
           in
             let xrefs' = ref [] in
+          if !read_debug then
+            Printf.eprintf
+              "****** Begin iteration: %i\n%!" (length starts_and_lens);
             iter
               (fun (start, len) ->
                 let these_xrefs =
@@ -1219,7 +1237,8 @@ let read_xref_stream i =
                   xrefs := drop !xrefs len;
                   let objnumber = ref start in
                     iter
-                      (function
+                      (fun x ->
+                       match x with
                        | Valid (offset, gen) ->
                            xrefs' =| (!objnumber, XRefPlain (offset, gen));
                            incr objnumber
@@ -1231,7 +1250,7 @@ let read_xref_stream i =
                 starts_and_lens;
               i.seek_in original_pos;
               if !read_debug then
-                Printf.printf "***READ_XREF_STREAM final result was %i xrefs\n"
+                Printf.eprintf "***READ_XREF_STREAM final result was %i xrefs\n%!"
                 (length !xrefs');
               rev !xrefs', xrefstream_objectnumber
 
@@ -1241,14 +1260,42 @@ let get_object i xrefs n =
   let lexemes = lex_object i xrefs parse false n in
     snd (parse lexemes)
 
+let is_linearized i =
+  try
+    ignore (read_header i);
+    let lexemes = lex_dictionary i in
+      let _, parsed = parse ~failure_is_ok:true lexemes in
+        match Pdf.lookup_direct (Pdf.empty ()) "/Linearized" parsed with
+        | Some (Pdf.Integer _ | Pdf.Real _) -> true
+        | _ -> false
+  with
+    _ -> false
+
+exception Revisions of int
+
+exception BadRevision
+
 (* Read a PDF from a channel. If [opt], streams are read immediately into
-memory. *)
-let read_pdf user_pw owner_pw opt i =
+memory. Revision: 1 = first revision, 2 = second revision etc. max_int = latest
+revision (default). If revision = -1, the file is not read, but instead the
+exception Revisions x is raised, giving the number of revisions. *)
+let read_pdf ?revision user_pw owner_pw opt i =
+  if !read_debug then
+    Printf.eprintf "read_pdf, revision is %s\n%!"
+      (match revision with None -> "None" | Some x -> string_of_int x);
+  begin match revision with
+     Some x when x < 1 && x <> (-1) -> raise BadRevision
+   | _ -> ()
+  end;
+  let revisions = ref 1 in
+  let current_revision = ref 0 in
+  let was_linearized = is_linearized i in
+  i.seek_in 0;
   if !read_debug then
     begin
-      flprint "Start of read_pdf\n";
-      Printf.printf "%s\n" i.Pdfio.source;
-      flprint "opt is "; Printf.printf "%b" opt; flprint "\n"
+      Printf.eprintf "Start of read_pdf\n";
+      Printf.eprintf "%s\n" i.Pdfio.source;
+      Printf.eprintf "opt is "; Printf.eprintf "%b" opt; Printf.eprintf "\n"
     end;
   let postdeletes = ref [] in
   let object_stream_ids = null_hash () in
@@ -1295,22 +1342,35 @@ let read_pdf user_pw owner_pw opt i =
             (Pdf.PDFError (Pdf.input_pdferror i "Could not find xref pointer"))
       | xrefchars -> xref := int_of_string (implode xrefchars);
       end;
-      if !read_debug then flprint "Reading Cross-reference table\n";
+      if !read_debug then Printf.eprintf "Reading Cross-reference table\n";
       while not !got_all_xref_sections do
+        if !read_debug then Printf.eprintf "Reading xref section at %i\n" !xref;
         i.seek_in !xref;
+        incr current_revision;
         (* Distinguish between xref table and xref stream. *)
         dropwhite i;
         (* Read cross-reference table *)
-        if peek_char i = Some 'x'
-          then iter addref (read_xref i)
-          else
-            begin
-              let refs, objnumbertodelete = read_xref_stream i in
-                postdeletes := objnumbertodelete::!postdeletes;
-                iter addref refs
-            end;
-        (* It is now assumed that [i] is at the start of the trailer dictionary.
-        *)
+        let read_table skip =
+          if !read_debug then
+            Printf.eprintf "Reading table for revision %i, skip = %b\n"
+            !current_revision skip;
+          if peek_char i = Some 'x'
+            then
+              iter (if skip then (function _ -> ()) else addref) (read_xref i)
+            else
+              if not skip then
+                let refs, objnumbertodelete = read_xref_stream i in
+                  (postdeletes := objnumbertodelete::!postdeletes;
+                   iter addref refs)
+        in
+        begin match revision with
+          None -> read_table false
+        | Some r when r <= !current_revision -> read_table false
+        | _ ->
+            if !read_debug then Printf.eprintf "Skipping revision %i\n" !current_revision;
+            read_table true
+        end;
+        (* It is now assumed that [i] is at the start of the trailer dictionary.  *)
         let trailerdict_current =
           let lexemes =
             lex_object_at true i opt parse (lex_object i xrefs parse opt)
@@ -1331,24 +1391,50 @@ let read_pdf user_pw owner_pw opt i =
             begin match lookup "/XRefStm" trailerdict_current with
             | Some (Pdf.Integer n) ->
                 i.seek_in n;
-                let refs, objnumbertodelete = read_xref_stream i in
-                  postdeletes := objnumbertodelete::!postdeletes;
-                  iter addref refs;
+                let read_table () =
+                  if !read_debug then
+                    Printf.eprintf "Reading table for revision stm %i\n" !current_revision;
+                  let refs, objnumbertodelete = read_xref_stream i in
+                    postdeletes := objnumbertodelete::!postdeletes;
+                    iter addref refs
+                in
+                  begin match revision with
+                    None -> read_table ()
+                  | Some r when r <= !current_revision -> read_table ()
+                  | _ ->
+                      if !read_debug then
+                        Printf.eprintf "Skipping /XRefStm in revision %i\n" !current_revision
+                  end;
             | _ -> ()
             end;
             (* Is there another to do? *)
             match lookup "/Prev" trailerdict_current with
             | None -> set got_all_xref_sections
-            | Some (Pdf.Integer n) -> xref := n
+            | Some (Pdf.Integer n) ->
+                revisions := !revisions + 1;
+                xref := n
             | _ ->
                 raise (Pdf.PDFError (Pdf.input_pdferror i "Malformed trailer"))
           end;
       done;
+      if revision = Some (-1) then
+        (* If there are exactly two "revisions", and the file is linearized,
+        then we return 1, since there is only one real revision. If there are
+        more than 2, we assume it has been incrementally updated, is not really
+        "linearized" at all, and so we return !revisions - 1, (the original
+        linearized part counting for two). Thus, one can never accidently read
+        half the original linearized file. *)
+        let real_revisions =
+          if was_linearized then !revisions - 1 else !revisions
+        in
+          raise (Revisions real_revisions)
+      else
       if !read_debug then
-        Printf.printf "*** READ %i XREF entries\n" (Hashtbl.length xrefs);
+        Printf.eprintf "*** READ %i XREF entries\n" (Hashtbl.length xrefs);
       let root =
         match lookup "/Root" !trailerdict with
-        | Some (Pdf.Indirect i) -> i
+        | Some (Pdf.Indirect i) ->
+            i
         | None ->
             raise (Pdf.PDFError (Pdf.input_pdferror i "No /Root entry"))
         | _ ->
@@ -1360,7 +1446,7 @@ let read_pdf user_pw owner_pw opt i =
           | Some (XRefStream _) -> 0
           | None -> raise Not_found
         in
-        if !read_debug then flprint "Reading non-stream objects\n";
+        if !read_debug then Printf.eprintf "Reading non-stream objects\n";
         let objects_nonstream =
           let objnumbers = ref [] in
             xrefs_table_iter
@@ -1380,7 +1466,7 @@ let read_pdf user_pw owner_pw opt i =
                      fun o -> o, (ref Pdf.ToParse, getgen o))
                 !objnumbers
           in
-          if !read_debug then flprint "Reading stream objects\n";
+          if !read_debug then Printf.eprintf "Reading stream objects\n";
           let objects_stream =
            let streamones =
              map
@@ -1444,7 +1530,7 @@ let read_pdf user_pw owner_pw opt i =
                    let themap =
                      let t = Hashtbl.create 200 in
                        let groups =
-                         let sf = fun (_, s, _) (_, s', _) -> compare s s' in
+                         let sf = fun (_, s, _) (_, s', _) -> compare_i s s' in
                            collate sf (sort sf streamones)
                        in
                          iter
@@ -1468,15 +1554,15 @@ let read_pdf user_pw owner_pw opt i =
          in
           if !read_debug then
             begin
-              Printf.printf
+              Printf.eprintf
                 "There were %i nonstream objects\n" (length objects_nonstream);
-              Printf.printf
+              Printf.eprintf
                 "There were %i stream objects\n" (length objects_stream);
-              flprint "\n"
+              Printf.eprintf "\n"
             end;
           objects_stream, objects_nonstream, root, trailerdict
     in
-      if !read_debug then flprint "Finishing up...\n";
+      if !read_debug then Printf.eprintf "Finishing up...\n";
       let objects = objects_stream @ objects_nonstream in
         (* Fix Size entry and remove Prev, XRefStm, Filter, Index, W, Type,
         and DecodeParms *)
@@ -1498,8 +1584,26 @@ let read_pdf user_pw owner_pw opt i =
                Pdf.objects_of_list (Some (get_object i xrefs)) objects;
              Pdf.root = root;
              Pdf.trailerdict = trailerdict';
+             Pdf.was_linearized = was_linearized;
              Pdf.saved_encryption = None}
           in
+          (* Check for a /Version in the document catalog *)
+          begin match Pdf.lookup_direct pdf "/Version" (Pdf.lookup_obj pdf root) with
+            Some (Pdf.Name s) ->
+              let major, minor =
+                try
+                  read_header
+                    (Pdfio.input_of_string
+                       ("%PDF-" ^ String.sub s 1 (String.length s - 1)))
+                with
+                  e ->
+                    flprint (Printexc.to_string e);
+                    (pdf.Pdf.major, pdf.Pdf.minor)
+              in
+                pdf.Pdf.major <- major;
+                pdf.Pdf.minor <- minor
+          | _ -> ()
+          end;
             (* Delete items in !postdeletes - these are any xref streams, and
             object streams (if finished with).  This allows decryption to be
             performed without accidently trying to decrypt these streams -
@@ -1508,22 +1612,34 @@ let read_pdf user_pw owner_pw opt i =
             iter (Pdf.removeobj pdf) !postdeletes;
             (* Add stream_object_ids *)
             pdf.Pdf.objects.Pdf.object_stream_ids <- object_stream_ids;
-            if !read_debug then
+            (*if !read_debug then
               begin
-                flprint "Object stream IDs:\n";
+                Printf.eprintf "Object stream IDs:\n";
                 Hashtbl.iter
                   (fun o s ->
-                     Printf.printf "Object %i was in stream %i\n" o s)
+                     Printf.eprintf "Object %i was in stream %i\n" o s)
                   object_stream_ids
-              end;
+              end;*)
             if !read_debug then
               begin
-                flprint "Done.\n";
+                Printf.eprintf "Done.\n";
                 match Pdf.lookup_direct pdf "/Encrypt" pdf.Pdf.trailerdict with
-                | Some _ -> flprint "***File is encrypted\n" | _ -> ()
+                | Some _ -> Printf.eprintf "***File is encrypted\n" | _ -> ()
               end;
             pdf
 
+(* Read the number of revisions of the document, by performing a dummy read. For
+example, if this function returns 3, then appropriate values to pass to
+?revision in a subsequent call to read_pdf are 1, 2, and 3. The position of
+input after this runs is unspecified. On a malformed PDF, will raise whatever
+error [read_pdf] does. *)
+let revisions i =
+  try
+    ignore (read_pdf ~revision:(-1) None None false i);
+    assert false
+  with
+    Revisions n -> n
+
 (* Malformed file reading. *)
 
 (* We read all trailerdicts in the file, preferring entries in later
@@ -1628,22 +1744,27 @@ let read_malformed_pdf upw opw i =
         | _ ->
             raise (Pdf.PDFError (Pdf.input_pdferror i "Malformed /Root entry"))
       in
-        Printf.eprintf "Malformed PDF reconstruction succeeded!\n";
-        flush stderr;
-        {Pdf.major = major;
-         Pdf.minor = minor;
-         Pdf.root = root;
-         Pdf.objects = Pdf.objects_of_list None objects;
-         Pdf.trailerdict = Pdf.Dictionary trailerdict;
-         Pdf.saved_encryption = None}
+        i.Pdfio.seek_in 0;
+        let was_linearized = is_linearized i in
+          Printf.eprintf "Malformed PDF reconstruction succeeded!\n";
+          flush stderr;
+          {Pdf.major = major;
+           Pdf.minor = minor;
+           Pdf.root = root;
+           Pdf.objects = Pdf.objects_of_list None objects;
+           Pdf.trailerdict = Pdf.Dictionary trailerdict;
+           Pdf.was_linearized = was_linearized;
+           Pdf.saved_encryption = None}
     
-let read_pdf upw opw opt i =
-  try read_pdf upw opw opt i with
+let read_pdf revision upw opw opt i =
+  try read_pdf ?revision upw opw opt i with
   | Pdf.PDFError s as e
       when String.length s >= 10 && String.sub s 0 10 = "Encryption" ->
       (* If it failed due to encryption not supported or user password not
       right, the error should be passed up - it's not a malformed file. *)
       raise e
+  | BadRevision ->
+      raise (Pdf.PDFError "Revision number too low when reading PDF")
   | e ->
       try read_malformed_pdf upw opw i with e' ->
         raise
@@ -1657,26 +1778,26 @@ let read_pdf upw opw opt i =
                    (Printexc.to_string e'))))
 
 (* Read a PDF into memory, including its streams. *)
-let pdf_of_channel ?(source = "channel") upw opw ch =
-  read_pdf upw opw true (input_of_channel ~source ch) 
+let pdf_of_channel ?revision ?(source = "channel") upw opw ch =
+  read_pdf revision upw opw true (input_of_channel ~source ch) 
 
 (* Same, but delay reading of streams. *)
-let pdf_of_channel_lazy ?(source = "channel") upw opw ch =
-  read_pdf upw opw false (input_of_channel ~source ch)
+let pdf_of_channel_lazy ?revision ?(source = "channel") upw opw ch =
+  read_pdf revision upw opw false (input_of_channel ~source ch)
 
 (* Similarly for inputs. *)
-let pdf_of_input upw opw i =
-  read_pdf upw opw true i
+let pdf_of_input ?revision upw opw i =
+  read_pdf revision upw opw true i
 
 (* And lazy on inputs. *)
-let pdf_of_input_lazy upw opw i =
-  read_pdf upw opw false i
+let pdf_of_input_lazy ?revision upw opw i =
+  read_pdf revision upw opw false i
 
 (* Read a whole PDF file into memory. Closes file. *)
-let pdf_of_file upw opw f =
+let pdf_of_file ?revision upw opw f =
   try 
     let fh = open_in_bin f in
-      let pdf = pdf_of_channel ~source:f upw opw fh in
+      let pdf = pdf_of_channel ?revision ~source:f upw opw fh in
         close_in fh;
         pdf
   with
@@ -1687,9 +1808,9 @@ let what_encryption pdf =
   if Pdfcrypt.is_encrypted pdf then
     let crypt, _, _, _, _, _, _ = Pdfcrypt.get_encryption_values pdf in
       match crypt with
-      | Pdf.ARC4 (40, _) -> Some (Pdfwrite.PDF40bit)
-      | Pdf.ARC4 (128, _) -> Some (Pdfwrite.PDF128bit)
-      | Pdf.AESV2 | Pdf.AESV3 _ ->
+      | Pdfcryptprimitives.ARC4 (40, _) -> Some (Pdfwrite.PDF40bit)
+      | Pdfcryptprimitives.ARC4 (128, _) -> Some (Pdfwrite.PDF128bit)
+      | Pdfcryptprimitives.AESV2 | Pdfcryptprimitives.AESV3 _ ->
           let metadata =
             match Pdf.lookup_direct pdf "/Encrypt" pdf.Pdf.trailerdict with
             | Some encrypt_dict ->
@@ -1702,9 +1823,9 @@ let what_encryption pdf =
             | _ -> assert false
           in
             begin match crypt with
-            | Pdf.AESV2 -> Some (Pdfwrite.AES128bit metadata)
-            | Pdf.AESV3 false -> Some (Pdfwrite.AES256bit metadata)
-            | Pdf.AESV3 true -> Some (Pdfwrite.AES256bitISO metadata)
+            | Pdfcryptprimitives.AESV2 -> Some (Pdfwrite.AES128bit metadata)
+            | Pdfcryptprimitives.AESV3 false -> Some (Pdfwrite.AES256bit metadata)
+            | Pdfcryptprimitives.AESV3 true -> Some (Pdfwrite.AES256bitISO metadata)
             | _ -> assert false
             end
       | _ -> None
@@ -1718,14 +1839,3 @@ let permissions pdf =
   else
     []
 
-let is_linearized i =
-  try
-    ignore (read_header i);
-    let lexemes = lex_dictionary i in
-      let _, parsed = parse ~failure_is_ok:true lexemes in
-        match Pdf.lookup_direct (Pdf.empty ()) "/Linearized" parsed with
-        | Some (Pdf.Integer _ | Pdf.Real _) -> true
-        | _ -> false
-  with
-    _ -> false
-
