@@ -21,6 +21,22 @@ let string_replace_all x x' s =
         done;
         Buffer.contents output
 
+let string_replace_all_lazy x x' s =
+  if x = "" then s else
+    let p = ref 0
+    and slen = String.length s
+    and xlen = String.length x in
+      let output = Buffer.create (slen * 2) in
+        while !p < slen do
+          try
+            if String.sub s !p xlen = x
+              then (Buffer.add_string output (x' ()); p := !p + xlen)
+              else (Buffer.add_char output s.[!p]; incr p)
+          with
+            _ -> Buffer.add_char output s.[!p]; incr p
+        done;
+        Buffer.contents output
+
 (* Print something and then flush standard output. *)
 let flprint s =
   print_string s; flush stdout
@@ -656,7 +672,7 @@ let rec lose_inner prev p = function
 let lose p = lose_inner [] p
 
 (* Make a list of length [n] with each element equal to [x]. *)
-let many x n =
+let many x ~n =
   Array.to_list (Array.make n x)
 
 (* A version where we need to apply unit each time, for instance when producing
@@ -817,7 +833,8 @@ let splitat points l =
 (* Select the nth element in a list (first is element 1) *)
 let select n l =
   try hd (drop l (n - 1)) with
-    Invalid_argument "drop" | Failure "hd" -> raise (Invalid_argument "select")
+    Invalid_argument _ (*"drop"*)
+  | Failure _ (*"hd"*) -> raise (Invalid_argument "select")
 
 (* Replace the nth element of a list (first is element 1) *)
 let rec replace_number_inner prev n e = function
