@@ -38,6 +38,8 @@ val string_of_char : char -> string
 in [s], returning a new string. *)
 val string_replace_all : string -> string -> string -> string
 
+val string_replace_all_lazy : string -> (unit -> string) -> string -> string
+
 (** {2 Lists} *)
 
 val hd : ('a list -> 'a)
@@ -233,7 +235,7 @@ matching a predicate. *)
 val lose : ('a -> bool) -> ('a list -> 'a list)
 
 (** [many x n] makes a list of length [n] with each element equal to [x]. *)
-val many : 'a -> int -> 'a list
+val many : 'a -> n:int -> 'a list
 
 (** A version where we need to apply unit each time, for instance when producing
 a list of random numbers. Result is ordered. *)
