@@ -11,18 +11,12 @@ let header pdf =
     pdf.Pdf.major
     pdf.Pdf.minor
 
-(* Build a cross-reference table string. (With new GADT printf, this may be
-faster with printf -- check). *)
-let pad_to_ten ch s =
-  let l = String.length s in
-    if l > 10 then raise (Pdf.PDFError "xref too big") else
-      let t = String.make 10 ch in
-        String.blit s 0 t (10 - l) l;
-        t
-
 let output_string_of_xref i n =
-  i.output_string (pad_to_ten '0' (string_of_int n));
-  i.output_string " 00000 n \n" 
+  let s = string_of_int n in
+    let l = String.length s in
+      for x = 0 to 10 - l - 1 do i.output_char '0' done;
+      i.output_string s;
+      i.output_string " 00000 n \n" 
 
 (* Write the cross-reference table to a channel. *)
 let write_xrefs xrefs i =
@@ -205,9 +199,10 @@ let string_of_pdf s =
     s;
   Buffer.contents b
 
-(* Inter-module recursion. *)
+(* Inter-module recursion, for debug *)
 let _ = Pdfcrypt.string_of_pdf := string_of_pdf
 let _ = Pdfcodec.string_of_pdf := string_of_pdf
+let _ = Pdf.string_of_pdf := string_of_pdf
 
 let debug_whole_pdf pdf =
   Printf.printf "trailerdict = %s\n" (string_of_pdf pdf.Pdf.trailerdict);
@@ -600,7 +595,11 @@ let pdf_to_output
       Hashtbl.length pdf.Pdf.objects.Pdf.object_stream_ids > 0
     then
       (reinstate_object_streams
-        compress_objstm (match encrypt with Some _ -> true | _ -> false) pdf, true)
+        compress_objstm
+        (match encrypt with Some _ -> true | _ ->
+          match recrypt with Some _ -> true | _ -> false)
+        pdf,
+        true)
     else
       ([], false) (* Weren't asked to preserve, or nothing to put in streams *)
   in
